@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=2 sw=2 et tw=78: */
 /* ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
  *
@@ -44,6 +45,13 @@
 
 #include "prio.h"
 #include "prprf.h"
+#if defined(XP_WIN)
+#include <windows.h>
+#elif defined(XP_MACOSX)
+#include <Carbon/Carbon.h>
+#elif defined(MOZ_WIDGET_GTK2)
+#include <gtk/gtk.h>
+#endif
 
 #include "nsAppDirectoryServiceDefs.h"
 #include "nsArrayEnumerator.h"
@@ -63,9 +71,9 @@
 #include "nsWidgetsCID.h"
 #include "nsXPIDLString.h"
 #include "nsXULAppAPI.h"
+#include "nsTextFormatter.h"
 
 #include "nsIAtom.h"
-#include "nsIBindingManager.h"
 #include "nsICommandLine.h"
 #include "nsICSSLoader.h"
 #include "nsICSSStyleSheet.h"
@@ -82,6 +90,7 @@
 #include "nsIFileChannel.h"
 #include "nsIFileURL.h"
 #include "nsIIOService.h"
+#include "nsIJARProtocolHandler.h"
 #include "nsIJARURI.h"
 #include "nsILocalFile.h"
 #include "nsILocaleService.h"
@@ -91,6 +100,8 @@
 #include "nsIPrefBranch.h"
 #include "nsIPrefBranch2.h"
 #include "nsIPresShell.h"
+#include "nsIProtocolHandler.h"
+#include "nsIResProtocolHandler.h"
 #include "nsIScriptError.h"
 #include "nsIServiceManager.h"
 #include "nsISimpleEnumerator.h"
@@ -101,7 +112,9 @@
 #include "nsIXPConnect.h"
 #include "nsIXULAppInfo.h"
 #include "nsIXULRuntime.h"
+#include "nsPresShellIterator.h"
 
+#ifdef MOZ_XUL
 // keep all the RDF stuff together, in case we can remove it in the far future
 #include "rdf.h"
 #include "nsRDFCID.h"
@@ -115,6 +128,15 @@
 #include "nsIRDFContainer.h"
 #include "nsIRDFContainerUtils.h"
 
+#define CHROME_URI "http://www.mozilla.org/rdf/chrome#"
+
+DEFINE_RDF_VOCAB(CHROME_URI, CHROME, packages);
+DEFINE_RDF_VOCAB(CHROME_URI, CHROME, package);
+DEFINE_RDF_VOCAB(CHROME_URI, CHROME, name);
+DEFINE_RDF_VOCAB(CHROME_URI, CHROME, platformPackage);
+
+#endif
+
 #define UILOCALE_CMD_LINE_ARG "UILocale"
 
 #define MATCH_OS_LOCALE_PREF "intl.locale.matchOS"
@@ -126,13 +148,6 @@ static NS_DEFINE_CID(kLookAndFeelCID, NS_LOOKANDFEEL_CID);
 
 nsChromeRegistry* nsChromeRegistry::gChromeRegistry;
 
-#define CHROME_URI "http://www.mozilla.org/rdf/chrome#"
-
-DEFINE_RDF_VOCAB(CHROME_URI, CHROME, packages);
-DEFINE_RDF_VOCAB(CHROME_URI, CHROME, package);
-DEFINE_RDF_VOCAB(CHROME_URI, CHROME, name);
-DEFINE_RDF_VOCAB(CHROME_URI, CHROME, platformPackage);
-
 ////////////////////////////////////////////////////////////////////////////////
 
 static void
@@ -248,7 +263,7 @@ nsChromeRegistry::nsProviderArray::GetProvider(const nsACString& aPreferred, Mat
   ProviderEntry* entry;
 
   while (i--) {
-    entry = NS_REINTERPRET_CAST(ProviderEntry*, mArray[i]);
+    entry = reinterpret_cast<ProviderEntry*>(mArray[i]);
     if (aPreferred.Equals(entry->provider))
       return entry;
 
@@ -315,7 +330,7 @@ nsChromeRegistry::nsProviderArray::EnumerateToArray(nsCStringArray *a)
 {
   PRInt32 i = mArray.Count();
   while (i--) {
-    ProviderEntry *entry = NS_REINTERPRET_CAST(ProviderEntry*, mArray[i]);
+    ProviderEntry *entry = reinterpret_cast<ProviderEntry*>(mArray[i]);
     a->AppendCString(entry->provider);
   }
 }
@@ -325,7 +340,7 @@ nsChromeRegistry::nsProviderArray::Clear()
 {
   PRInt32 i = mArray.Count();
   while (i--) {
-    ProviderEntry* entry = NS_REINTERPRET_CAST(ProviderEntry*, mArray[i]);
+    ProviderEntry* entry = reinterpret_cast<ProviderEntry*>(mArray[i]);
     delete entry;
   }
 
@@ -337,17 +352,10 @@ nsChromeRegistry::PackageEntry::PackageEntry(const nsACString& aPackage) :
 {
 }
 
-const void*
-nsChromeRegistry::GetKey(PLDHashTable *table, PLDHashEntryHdr *entry)
-{
-  PackageEntry* pentry = NS_STATIC_CAST(PackageEntry*, entry);
-  return (nsACString*) &pentry->package;
-}
-
 PLHashNumber
 nsChromeRegistry::HashKey(PLDHashTable *table, const void *key)
 {
-  const nsACString& str = *NS_REINTERPRET_CAST(const nsACString*, key);
+  const nsACString& str = *reinterpret_cast<const nsACString*>(key);
   return HashString(str);
 }
 
@@ -355,15 +363,15 @@ PRBool
 nsChromeRegistry::MatchKey(PLDHashTable *table, const PLDHashEntryHdr *entry,
                            const void *key)
 {
-  const nsACString& str = *NS_REINTERPRET_CAST(const nsACString*, key);
-  const PackageEntry* pentry = NS_STATIC_CAST(const PackageEntry*, entry);
+  const nsACString& str = *reinterpret_cast<const nsACString*>(key);
+  const PackageEntry* pentry = static_cast<const PackageEntry*>(entry);
   return str.Equals(pentry->package);
 }
 
 void
 nsChromeRegistry::ClearEntry(PLDHashTable *table, PLDHashEntryHdr *entry)
 {
-  PackageEntry* pentry = NS_STATIC_CAST(PackageEntry*, entry);
+  PackageEntry* pentry = static_cast<PackageEntry*>(entry);
   pentry->~PackageEntry();
 }
 
@@ -371,7 +379,7 @@ PRBool
 nsChromeRegistry::InitEntry(PLDHashTable *table, PLDHashEntryHdr *entry,
                             const void *key)
 {
-  const nsACString& str = *NS_REINTERPRET_CAST(const nsACString*, key);
+  const nsACString& str = *reinterpret_cast<const nsACString*>(key);
 
   new (entry) PackageEntry(str);
   return PR_TRUE;
@@ -381,7 +389,6 @@ const PLDHashTableOps
 nsChromeRegistry::kTableOps = {
   PL_DHashAllocTable,
   PL_DHashFreeTable,
-  GetKey,
   HashKey,
   MatchKey,
   PL_DHashMoveEntryStub,
@@ -423,17 +430,25 @@ nsChromeRegistry::OverlayListHash::GetArray(nsIURI* aBase)
 
 nsChromeRegistry::~nsChromeRegistry()
 {
-  PL_DHashTableFinish(&mPackagesHash);
+  if (mPackagesHash.ops)
+    PL_DHashTableFinish(&mPackagesHash);
   gChromeRegistry = nsnull;
 }
 
-NS_IMPL_THREADSAFE_ISUPPORTS6(nsChromeRegistry,
-                              nsIChromeRegistry,
-                              nsIXULChromeRegistry,
-                              nsIToolkitChromeRegistry,
-                              nsIXULOverlayProvider,
-                              nsIObserver,
-                              nsISupportsWeakReference)
+NS_INTERFACE_MAP_BEGIN(nsChromeRegistry)
+  NS_INTERFACE_MAP_ENTRY(nsIChromeRegistry)
+  NS_INTERFACE_MAP_ENTRY(nsIXULChromeRegistry)
+  NS_INTERFACE_MAP_ENTRY(nsIToolkitChromeRegistry)
+#ifdef MOZ_XUL
+  NS_INTERFACE_MAP_ENTRY(nsIXULOverlayProvider)
+#endif
+  NS_INTERFACE_MAP_ENTRY(nsIObserver)
+  NS_INTERFACE_MAP_ENTRY(nsISupportsWeakReference)
+  NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsIChromeRegistry)
+NS_INTERFACE_MAP_END
+
+NS_IMPL_ADDREF(nsChromeRegistry)
+NS_IMPL_RELEASE(nsChromeRegistry)
 
 ////////////////////////////////////////////////////////////////////////////////
 // nsIChromeRegistry methods:
@@ -474,6 +489,22 @@ nsChromeRegistry::Init()
 
   NS_RegisterStaticAtoms(atoms, NS_ARRAY_LENGTH(atoms));
   
+  // Check to see if necko and the JAR protocol handler are registered yet
+  // if not, somebody is doing work during XPCOM registration that they
+  // shouldn't be doing. See bug 292549, where JS components are trying
+  // to call Components.utils.import("chrome:///") early in registration
+
+  nsCOMPtr<nsIIOService> io (do_GetIOService());
+  if (!io) return NS_ERROR_FAILURE;
+
+  nsCOMPtr<nsIProtocolHandler> ph;
+  rv = io->GetProtocolHandler("jar", getter_AddRefs(ph));
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  nsCOMPtr<nsIJARProtocolHandler> jph = do_QueryInterface(ph);
+  if (!jph)
+    return NS_ERROR_NOT_INITIALIZED;
+
   if (!PL_DHashTableInit(&mPackagesHash, &kTableOps,
                          nsnull, sizeof(PackageEntry), 16))
     return NS_ERROR_FAILURE;
@@ -666,13 +697,32 @@ nsChromeRegistry::Canonify(nsIURL* aChromeURL)
     aChromeURL->SetPath(path);
   }
   else {
-    nsCAutoString filePath;
-    rv = aChromeURL->GetFilePath(filePath);
-    NS_ENSURE_SUCCESS(rv, rv);
-
-    if (filePath.Find(NS_LITERAL_CSTRING("..")) != -1 ||
-        filePath.FindChar(':') != -1) {
-      return NS_ERROR_DOM_BAD_URI;
+    // prevent directory traversals ("..")
+    // path is already unescaped once, but uris can get unescaped twice
+    const char* pos = path.BeginReading();
+    const char* end = path.EndReading();
+    while (pos < end) {
+      switch (*pos) {
+        case ':':
+          return NS_ERROR_DOM_BAD_URI;
+        case '.':
+          if (pos[1] == '.')
+            return NS_ERROR_DOM_BAD_URI;
+          break;
+        case '%':
+          // chrome: URIs with double-escapes are trying to trick us.
+          // watch for %2e, and %25 in case someone triple unescapes
+          if (pos[1] == '2' &&
+               ( pos[2] == 'e' || pos[2] == 'E' || 
+                 pos[2] == '5' ))
+            return NS_ERROR_DOM_BAD_URI;
+          break;
+        case '?':
+        case '#':
+          pos = end;
+          continue;
+      }
+      ++pos;
     }
   }
 
@@ -699,7 +749,7 @@ nsChromeRegistry::ConvertChromeURL(nsIURI* aChromeURI, nsIURI* *aResult)
   NS_ENSURE_SUCCESS(rv, rv);
 
   PackageEntry* entry =
-    NS_STATIC_CAST(PackageEntry*, PL_DHashTableOperate(&mPackagesHash,
+    static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
                                                        & (nsACString&) package,
                                                        PL_DHASH_LOOKUP));
 
@@ -707,7 +757,7 @@ nsChromeRegistry::ConvertChromeURL(nsIURI* aChromeURI, nsIURI* *aResult)
     if (!mInitialized)
       return NS_ERROR_NOT_INITIALIZED;
 
-    LogMessage("No chrome package registered for chrome://%s/%s/%s .",
+    LogMessage("No chrome package registered for chrome://%s/%s/%s",
                package.get(), provider.get(), path.get());
 
     return NS_ERROR_FAILURE;
@@ -730,12 +780,12 @@ nsChromeRegistry::ConvertChromeURL(nsIURI* aChromeURI, nsIURI* *aResult)
   else if (provider.EqualsLiteral("skin")) {
     baseURI = entry->skins.GetBase(mSelectedSkin, nsProviderArray::ANY);
   }
-  else {
+  else if (provider.EqualsLiteral("content")) {
     baseURI = entry->baseURI;
   }
 
   if (!baseURI) {
-    LogMessage("No chrome package registered for chrome://%s/%s/%s .",
+    LogMessage("No chrome package registered for chrome://%s/%s/%s",
                package.get(), provider.get(), path.get());
     return NS_ERROR_FAILURE;
   }
@@ -747,7 +797,7 @@ nsresult
 nsChromeRegistry::GetSelectedLocale(const nsACString& aPackage, nsACString& aLocale)
 {
   PackageEntry* entry =
-    NS_STATIC_CAST(PackageEntry*, PL_DHashTableOperate(&mPackagesHash,
+    static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
                                                        & aPackage,
                                                        PL_DHASH_LOOKUP));
 
@@ -770,7 +820,7 @@ nsChromeRegistry::GetLocalesForPackage(const nsACString& aPackage,
     return NS_ERROR_OUT_OF_MEMORY;
 
   PackageEntry* entry =
-    NS_STATIC_CAST(PackageEntry*, PL_DHashTableOperate(&mPackagesHash,
+    static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
                                                        & aPackage,
                                                        PL_DHASH_LOOKUP));
 
@@ -785,6 +835,7 @@ nsChromeRegistry::GetLocalesForPackage(const nsACString& aPackage,
   return rv;
 }
 
+#ifdef MOZ_XUL
 NS_IMETHODIMP
 nsChromeRegistry::GetStyleOverlays(nsIURI *aChromeURL,
                                    nsISimpleEnumerator **aResult)
@@ -805,6 +856,7 @@ nsChromeRegistry::GetXULOverlays(nsIURI *aChromeURL, nsISimpleEnumerator **aResu
 
   return NS_NewArrayEnumerator(aResult, *parray);
 }
+#endif // MOZ_XUL
 
 ////////////////////////////////////////////////////////////////////////
 
@@ -824,12 +876,16 @@ static void FlushSkinBindingsForWindow(nsIDOMWindowInternal* aWindow)
     return;
 
   // Annihilate all XBL bindings.
-  document->BindingManager()->FlushSkinBindings();
+  document->FlushSkinBindings();
 }
 
 // XXXbsmedberg: move this to nsIWindowMediator
 NS_IMETHODIMP nsChromeRegistry::RefreshSkins()
 {
+  nsCOMPtr<nsICSSLoader> cssLoader(do_CreateInstance(kCSSLoaderCID));
+  if (!cssLoader)
+    return NS_OK;
+
   nsCOMPtr<nsIWindowMediator> windowMediator
     (do_GetService(NS_WINDOWMEDIATOR_CONTRACTID));
   if (!windowMediator)
@@ -860,7 +916,7 @@ NS_IMETHODIMP nsChromeRegistry::RefreshSkins()
     if (protoWindow) {
       nsCOMPtr<nsIDOMWindowInternal> domWindow = do_QueryInterface(protoWindow);
       if (domWindow)
-        RefreshWindow(domWindow);
+        RefreshWindow(domWindow, cssLoader);
     }
     windowEnumerator->HasMoreElements(&more);
   }
@@ -875,7 +931,7 @@ nsChromeRegistry::FlushSkinCaches()
     do_GetService("@mozilla.org/observer-service;1");
   NS_ASSERTION(obsSvc, "Couldn't get observer service.");
 
-  obsSvc->NotifyObservers(NS_STATIC_CAST(nsIChromeRegistry*, this),
+  obsSvc->NotifyObservers(static_cast<nsIChromeRegistry*>(this),
                           NS_CHROME_FLUSH_SKINS_TOPIC, nsnull);
 }
 
@@ -888,7 +944,8 @@ static PRBool IsChromeURI(nsIURI* aURI)
 }
 
 // XXXbsmedberg: move this to windowmediator
-nsresult nsChromeRegistry::RefreshWindow(nsIDOMWindowInternal* aWindow)
+nsresult nsChromeRegistry::RefreshWindow(nsIDOMWindowInternal* aWindow,
+                                         nsICSSLoader* aCSSLoader)
 {
   // Deal with our subframes first.
   nsCOMPtr<nsIDOMWindowCollection> frames;
@@ -900,7 +957,7 @@ nsresult nsChromeRegistry::RefreshWindow(nsIDOMWindowInternal* aWindow)
     nsCOMPtr<nsIDOMWindow> childWin;
     frames->Item(j, getter_AddRefs(childWin));
     nsCOMPtr<nsIDOMWindowInternal> childInt(do_QueryInterface(childWin));
-    RefreshWindow(childInt);
+    RefreshWindow(childInt, aCSSLoader);
   }
 
   nsresult rv;
@@ -915,10 +972,9 @@ nsresult nsChromeRegistry::RefreshWindow(nsIDOMWindowInternal* aWindow)
     return NS_OK;
 
   // Deal with the agent sheets first.  Have to do all the style sets by hand.
-  PRUint32 shellCount = document->GetNumberOfShells();
-  for (PRUint32 k = 0; k < shellCount; k++) {
-    nsIPresShell *shell = document->GetShellAt(k);
-
+  nsPresShellIterator iter(document);
+  nsCOMPtr<nsIPresShell> shell;
+  while ((shell = iter.GetNextShell())) {
     // Reload only the chrome URL agent style sheets.
     nsCOMArray<nsIStyleSheet> agentSheets;
     rv = shell->GetAgentStyleSheets(agentSheets);
@@ -935,7 +991,7 @@ nsresult nsChromeRegistry::RefreshWindow(nsIDOMWindowInternal* aWindow)
       if (IsChromeURI(uri)) {
         // Reload the sheet.
         nsCOMPtr<nsICSSStyleSheet> newSheet;
-        rv = LoadStyleSheetWithURL(uri, PR_TRUE, getter_AddRefs(newSheet));
+        rv = aCSSLoader->LoadSheetSync(uri, PR_TRUE, getter_AddRefs(newSheet));
         if (NS_FAILED(rv)) return rv;
         if (newSheet) {
           rv = newAgentSheets.AppendObject(newSheet) ? NS_OK : NS_ERROR_FAILURE;
@@ -987,7 +1043,7 @@ nsresult nsChromeRegistry::RefreshWindow(nsIDOMWindowInternal* aWindow)
       // XXX what about chrome sheets that have a title or are disabled?  This
       // only works by sheer dumb luck.
       // XXXbz this should really use the document's CSSLoader!
-      LoadStyleSheetWithURL(uri, PR_FALSE, getter_AddRefs(newSheet));
+      aCSSLoader->LoadSheetSync(uri, getter_AddRefs(newSheet));
       // Even if it's null, we put in in there.
       newSheets.AppendObject(newSheet);
     }
@@ -1056,17 +1112,6 @@ nsChromeRegistry::ReloadChrome()
   return rv;
 }
 
-nsresult
-nsChromeRegistry::LoadStyleSheetWithURL(nsIURI* aURL, PRBool aEnableUnsafeRules, nsICSSStyleSheet** aSheet)
-{
-  *aSheet = nsnull;
-
-  nsCOMPtr<nsICSSLoader_MOZILLA_1_8_BRANCH> cssLoader = do_GetService(kCSSLoaderCID);
-  if (!cssLoader) return NS_ERROR_FAILURE;
-
-  return cssLoader->LoadSheetSync(aURL, aEnableUnsafeRules, aSheet);
-}
-
 NS_IMETHODIMP
 nsChromeRegistry::AllowScriptsForPackage(nsIURI* aChromeURI, PRBool *aResult)
 {
@@ -1092,6 +1137,40 @@ nsChromeRegistry::AllowScriptsForPackage(nsIURI* aChromeURI, PRBool *aResult)
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsChromeRegistry::AllowContentToAccess(nsIURI *aURI, PRBool *aResult)
+{
+  nsresult rv;
+
+  *aResult = PR_FALSE;
+
+#ifdef DEBUG
+  PRBool isChrome;
+  aURI->SchemeIs("chrome", &isChrome);
+  NS_ASSERTION(isChrome, "Non-chrome URI passed to AllowContentToAccess!");
+#endif
+
+  nsCOMPtr<nsIURL> url = do_QueryInterface(aURI);
+  if (!url) {
+    NS_ERROR("Chrome URL doesn't implement nsIURL.");
+    return NS_ERROR_UNEXPECTED;
+  }
+
+  nsCAutoString package;
+  rv = url->GetHostPort(package);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  PackageEntry *entry =
+    static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
+                                                    & (nsACString&) package,
+                                                    PL_DHASH_LOOKUP));
+
+  if (PL_DHASH_ENTRY_IS_BUSY(entry)) {
+    *aResult = !!(entry->flags & PackageEntry::CONTENT_ACCESSIBLE);
+  }
+  return NS_OK;
+}
+
 static PLDHashOperator
 RemoveAll(PLDHashTable *table, PLDHashEntryHdr *entry, PRUint32 number, void *arg)
 {
@@ -1118,6 +1197,7 @@ nsChromeRegistry::CheckForNewChrome()
 
   nsCOMPtr<nsIFileURL> manifestFileURL (do_QueryInterface(manifestURI));
   NS_ASSERTION(manifestFileURL, "Not a nsIFileURL!");
+  NS_ENSURE_TRUE(manifestFileURL, NS_ERROR_UNEXPECTED);
 
   nsCOMPtr<nsIFile> manifest;
   manifestFileURL->GetFile(getter_AddRefs(manifest));
@@ -1148,8 +1228,12 @@ nsChromeRegistry::CheckForNewChrome()
 
     linstalled->SetNativeLeafName(NS_LITERAL_CSTRING("installed-chrome.txt"));
 
+    rv = linstalled->Exists(&exists);
+    NS_ENSURE_SUCCESS(rv, rv);
+
     // process installed-chrome.txt into app-chrome.manifest
-    ProcessNewChromeFile(linstalled, manifestURI);
+    if (exists)
+      ProcessNewChromeFile(linstalled, manifestURI);
   }
 
   nsCOMPtr<nsIProperties> dirSvc (do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
@@ -1411,6 +1495,7 @@ NS_IMETHODIMP nsChromeRegistry::Observe(nsISupports *aSubject, const char *aTopi
   return rv;
 }
 
+#ifdef MOZ_XUL
 static nsresult
 GetContainerEnumerator(nsIRDFDataSource* ds, nsIRDFResource* res,
                        nsISimpleEnumerator* *aResult, PRInt32 *aCountResult = nsnull)
@@ -1808,6 +1893,18 @@ nsChromeRegistry::ProcessOverlays(PRFileDesc *fd, nsIRDFDataSource* aDS,
   }
 }
 
+#else // MOZ_XUL
+
+NS_IMETHODIMP
+nsChromeRegistry::ProcessContentsManifest(nsIURI* aOldManifest, nsIURI* aFile,
+                                          nsIURI* aBaseURI, PRBool aAppend,
+                                          PRBool aSkinOnly)
+{
+  return NS_ERROR_NOT_IMPLEMENTED;
+}
+
+#endif // MOZ_XUL
+
 nsresult
 nsChromeRegistry::ProcessManifest(nsILocalFile* aManifest, PRBool aSkinOnly)
 {
@@ -1906,6 +2003,7 @@ enum TriState {
 /**
  * Check for a modifier flag of the following form:
  *   "flag=string"
+ *   "flag!=string"
  * @param aFlag The flag to compare.
  * @param aData The tokenized data to check; this is lowercased
  *              before being passed in.
@@ -1924,15 +2022,22 @@ CheckStringFlag(const nsSubstring& aFlag, const nsSubstring& aData,
   if (!StringBeginsWith(aData, aFlag))
     return PR_FALSE;
 
-  if (aData[aFlag.Length()] != '=')
-    return PR_FALSE;
+  PRBool comparison = PR_TRUE;
+  if (aData[aFlag.Length()] != '=') {
+    if (aData[aFlag.Length()] == '!' &&
+        aData.Length() >= aFlag.Length() + 2 &&
+        aData[aFlag.Length() + 1] == '=')
+      comparison = PR_FALSE;
+    else
+      return PR_FALSE;
+  }
 
   if (aResult != eOK) {
-    nsDependentSubstring testdata = Substring(aData, aFlag.Length() + 1);
+    nsDependentSubstring testdata = Substring(aData, aFlag.Length() + (comparison ? 1 : 2));
     if (testdata.Equals(aValue))
-      aResult = eOK;
+      aResult = comparison ? eOK : eBad;
     else
-      aResult = eBad;
+      aResult = comparison ? eBad : eOK;
   }
 
   return PR_TRUE;
@@ -1948,7 +2053,8 @@ CheckStringFlag(const nsSubstring& aFlag, const nsSubstring& aData,
  * @param aFlag The flag to compare.
  * @param aData The tokenized data to check; this is lowercased
  *              before being passed in.
- * @param aValue The value that is expected.
+ * @param aValue The value that is expected. If this is empty then no
+ *               comparison will match.
  * @param aChecker the version checker to use. If null, aResult will always
  *                 be eBad.
  * @param aResult If this is eOK when passed in, this is left alone.
@@ -1971,6 +2077,12 @@ CheckVersionFlag(const nsSubstring& aFlag, const nsSubstring& aData,
   if (!StringBeginsWith(aData, aFlag))
     return PR_FALSE;
 
+  if (aValue.Length() == 0) {
+    if (aResult != eOK)
+      aResult = eBad;
+    return PR_TRUE;
+  }
+
   PRUint32 comparison;
   nsAutoString testdata;
 
@@ -2034,6 +2146,16 @@ CheckVersionFlag(const nsSubstring& aFlag, const nsSubstring& aData,
   return PR_TRUE;
 }
 
+static void
+EnsureLowerCase(char *aBuf)
+{
+  for (; *aBuf; ++aBuf) {
+    char ch = *aBuf;
+    if (ch >= 'A' && ch <= 'Z')
+      *aBuf = ch + 'a' - 'A';
+  }
+}
+
 nsresult
 nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
                                         nsILocalFile* aManifest,
@@ -2043,12 +2165,22 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
   NS_NAMED_LITERAL_STRING(kPlatform, "platform");
   NS_NAMED_LITERAL_STRING(kXPCNativeWrappers, "xpcnativewrappers");
+  NS_NAMED_LITERAL_STRING(kContentAccessible, "contentaccessible");
   NS_NAMED_LITERAL_STRING(kApplication, "application");
   NS_NAMED_LITERAL_STRING(kAppVersion, "appversion");
+  NS_NAMED_LITERAL_STRING(kOs, "os");
+  NS_NAMED_LITERAL_STRING(kOsVersion, "osversion");
 
   nsCOMPtr<nsIIOService> io (do_GetIOService());
   if (!io) return NS_ERROR_FAILURE;
 
+  nsCOMPtr<nsIProtocolHandler> ph;
+  rv = io->GetProtocolHandler("resource", getter_AddRefs(ph));
+  NS_ENSURE_SUCCESS(rv, rv);
+  
+  nsCOMPtr<nsIResProtocolHandler> rph (do_QueryInterface(ph));
+  if (!rph) return NS_ERROR_FAILURE;
+
   nsCOMPtr<nsIURI> manifestURI;
   rv = io->NewFileURI(aManifest, getter_AddRefs(manifestURI));
   NS_ENSURE_SUCCESS(rv, rv);
@@ -2058,6 +2190,7 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
   nsAutoString appID;
   nsAutoString appVersion;
+  nsAutoString osTarget;
   nsCOMPtr<nsIXULAppInfo> xapp (do_GetService(XULAPPINFO_SERVICE_CONTRACTID));
   if (xapp) {
     nsCAutoString s;
@@ -2068,7 +2201,38 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
     rv = xapp->GetVersion(s);
     if (NS_SUCCEEDED(rv))
       CopyUTF8toUTF16(s, appVersion);
+    
+    nsCOMPtr<nsIXULRuntime> xruntime (do_QueryInterface(xapp));
+    if (xruntime) {
+      rv = xruntime->GetOS(s);
+      if (NS_SUCCEEDED(rv)) {
+        CopyUTF8toUTF16(s, osTarget);
+        ToLowerCase(osTarget);
+      }
+    }
   }
+  
+  nsAutoString osVersion;
+#if defined(XP_WIN)
+  OSVERSIONINFO info = { sizeof(OSVERSIONINFO) };
+  if (GetVersionEx(&info)) {
+    nsTextFormatter::ssprintf(osVersion, NS_LITERAL_STRING("%ld.%ld").get(),
+                                         info.dwMajorVersion,
+                                         info.dwMinorVersion);
+  }
+#elif defined(XP_MACOSX)
+  long majorVersion, minorVersion;
+  if ((Gestalt(gestaltSystemVersionMajor, &majorVersion) == noErr) &&
+      (Gestalt(gestaltSystemVersionMinor, &minorVersion) == noErr)) {
+    nsTextFormatter::ssprintf(osVersion, NS_LITERAL_STRING("%ld.%ld").get(),
+                                         majorVersion,
+                                         minorVersion);
+  }
+#elif defined(MOZ_WIDGET_GTK2)
+  nsTextFormatter::ssprintf(osVersion, NS_LITERAL_STRING("%ld.%ld").get(),
+                                       gtk_major_version,
+                                       gtk_minor_version);
+#endif
 
   char *token;
   char *newline = buf;
@@ -2099,13 +2263,18 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         continue;
       }
 
+      EnsureLowerCase(package);
+
       // NOTE: We check for platform and xpcnativewrappers modifiers on
       // content packages, but they are *applied* to content|skin|locale.
 
       PRBool platform = PR_FALSE;
       PRBool xpcNativeWrappers = PR_TRUE;
+      PRBool contentAccessible = PR_FALSE;
       TriState stAppVersion = eUnspecified;
       TriState stApp = eUnspecified;
+      TriState stOsVersion = eUnspecified;
+      TriState stOs = eUnspecified;
 
       PRBool badFlag = PR_FALSE;
 
@@ -2116,7 +2285,10 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
         if (CheckFlag(kPlatform, wtoken, platform) ||
             CheckFlag(kXPCNativeWrappers, wtoken, xpcNativeWrappers) ||
+            CheckFlag(kContentAccessible, wtoken, contentAccessible) ||
             CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
             CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
           continue;
 
@@ -2126,7 +2298,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         badFlag = PR_TRUE;
       }
 
-      if (badFlag || stApp == eBad || stAppVersion == eBad)
+      if (badFlag || stApp == eBad || stAppVersion == eBad || 
+          stOs == eBad || stOsVersion == eBad)
         continue;
 
       nsCOMPtr<nsIURI> resolved;
@@ -2136,7 +2309,7 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         continue;
 
       PackageEntry* entry =
-        NS_STATIC_CAST(PackageEntry*, PL_DHashTableOperate(&mPackagesHash,
+        static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
                                                             & (const nsACString&) nsDependentCString(package),
                                                             PL_DHASH_ADD));
       if (!entry)
@@ -2146,16 +2319,17 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
       if (platform)
         entry->flags |= PackageEntry::PLATFORM_PACKAGE;
-      if (xpcNativeWrappers) {
+      if (xpcNativeWrappers)
         entry->flags |= PackageEntry::XPCNATIVEWRAPPERS;
-        if (xpc) {
-          nsCAutoString urlp("chrome://");
-          urlp.Append(package);
-          urlp.Append('/');
+      if (contentAccessible)
+        entry->flags |= PackageEntry::CONTENT_ACCESSIBLE;
+      if (xpc) {
+        nsCAutoString urlp("chrome://");
+        urlp.Append(package);
+        urlp.Append('/');
 
-          rv = xpc->FlagSystemFilenamePrefix(urlp.get());
-          NS_ENSURE_SUCCESS(rv, rv);
-        }
+        rv = xpc->FlagSystemFilenamePrefix(urlp.get(), xpcNativeWrappers);
+        NS_ENSURE_SUCCESS(rv, rv);
       }
     }
     else if (!strcmp(token, "locale")) {
@@ -2173,8 +2347,12 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         continue;
       }
 
+      EnsureLowerCase(package);
+
       TriState stAppVersion = eUnspecified;
       TriState stApp = eUnspecified;
+      TriState stOs = eUnspecified;
+      TriState stOsVersion = eUnspecified;
 
       PRBool badFlag = PR_FALSE;
 
@@ -2184,6 +2362,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         ToLowerCase(wtoken);
 
         if (CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
             CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
           continue;
 
@@ -2193,7 +2373,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         badFlag = PR_TRUE;
       }
 
-      if (badFlag || stApp == eBad || stAppVersion == eBad)
+      if (badFlag || stApp == eBad || stAppVersion == eBad ||
+          stOs == eBad || stOsVersion == eBad)
         continue;
 
       nsCOMPtr<nsIURI> resolved;
@@ -2203,7 +2384,7 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         continue;
 
       PackageEntry* entry =
-        NS_STATIC_CAST(PackageEntry*, PL_DHashTableOperate(&mPackagesHash,
+        static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
                                                             & (const nsACString&) nsDependentCString(package),
                                                             PL_DHASH_ADD));
       if (!entry)
@@ -2221,8 +2402,12 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         continue;
       }
 
+      EnsureLowerCase(package);
+
       TriState stAppVersion = eUnspecified;
       TriState stApp = eUnspecified;
+      TriState stOs = eUnspecified;
+      TriState stOsVersion = eUnspecified;
 
       PRBool badFlag = PR_FALSE;
 
@@ -2232,6 +2417,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         ToLowerCase(wtoken);
 
         if (CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
             CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
           continue;
 
@@ -2241,7 +2428,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         badFlag = PR_TRUE;
       }
 
-      if (badFlag || stApp == eBad || stAppVersion == eBad)
+      if (badFlag || stApp == eBad || stAppVersion == eBad ||
+          stOs == eBad || stOsVersion == eBad)
         continue;
 
       nsCOMPtr<nsIURI> resolved;
@@ -2251,7 +2439,7 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         continue;
 
       PackageEntry* entry =
-        NS_STATIC_CAST(PackageEntry*, PL_DHashTableOperate(&mPackagesHash,
+        static_cast<PackageEntry*>(PL_DHashTableOperate(&mPackagesHash,
                                                             & (const nsACString&) nsDependentCString(package),
                                                             PL_DHASH_ADD));
       if (!entry)
@@ -2275,6 +2463,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
       TriState stAppVersion = eUnspecified;
       TriState stApp = eUnspecified;
+      TriState stOs = eUnspecified;
+      TriState stOsVersion = eUnspecified;
 
       PRBool badFlag = PR_FALSE;
 
@@ -2284,6 +2474,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         ToLowerCase(wtoken);
 
         if (CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
             CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
           continue;
 
@@ -2293,7 +2485,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         badFlag = PR_TRUE;
       }
 
-      if (badFlag || stApp == eBad || stAppVersion == eBad)
+      if (badFlag || stApp == eBad || stAppVersion == eBad ||
+          stOs == eBad || stOsVersion == eBad)
         continue;
 
       nsCOMPtr<nsIURI> baseuri, overlayuri;
@@ -2319,6 +2512,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
       TriState stAppVersion = eUnspecified;
       TriState stApp = eUnspecified;
+      TriState stOs = eUnspecified;
+      TriState stOsVersion = eUnspecified;
 
       PRBool badFlag = PR_FALSE;
 
@@ -2328,6 +2523,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         ToLowerCase(wtoken);
 
         if (CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
             CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
           continue;
 
@@ -2337,7 +2534,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         badFlag = PR_TRUE;
       }
 
-      if (badFlag || stApp == eBad || stAppVersion == eBad)
+      if (badFlag || stApp == eBad || stAppVersion == eBad ||
+          stOs == eBad || stOsVersion == eBad)
         continue;
 
       nsCOMPtr<nsIURI> baseuri, overlayuri;
@@ -2367,6 +2565,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
       TriState stAppVersion = eUnspecified;
       TriState stApp = eUnspecified;
+      TriState stOs = eUnspecified;
+      TriState stOsVersion = eUnspecified;
 
       PRBool badFlag = PR_FALSE;
 
@@ -2376,6 +2576,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         ToLowerCase(wtoken);
 
         if (CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
             CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
           continue;
 
@@ -2385,7 +2587,8 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
         badFlag = PR_TRUE;
       }
 
-      if (badFlag || stApp == eBad || stAppVersion == eBad)
+      if (badFlag || stApp == eBad || stAppVersion == eBad ||
+          stOs == eBad || stOsVersion == eBad)
         continue;
 
       nsCOMPtr<nsIURI> chromeuri, resolveduri;
@@ -2398,6 +2601,72 @@ nsChromeRegistry::ProcessManifestBuffer(char *buf, PRInt32 length,
 
       mOverrideTable.Put(chromeuri, resolveduri);
     }
+    else if (!strcmp(token, "resource")) {
+      if (aSkinOnly) {
+        LogMessageWithContext(manifestURI, line, nsIScriptError::warningFlag,
+                              "Warning: Ignoring resource registration in skin-only manifest.");
+        continue;
+      }
+
+      char *package = nsCRT::strtok(whitespace, kWhitespace, &whitespace);
+      char *uri     = nsCRT::strtok(whitespace, kWhitespace, &whitespace);
+      if (!package || !uri) {
+        LogMessageWithContext(manifestURI, line, nsIScriptError::warningFlag,
+                              "Warning: Malformed resource registration.");
+        continue;
+      }
+
+      EnsureLowerCase(package);
+
+      TriState stAppVersion = eUnspecified;
+      TriState stApp = eUnspecified;
+      TriState stOsVersion = eUnspecified;
+      TriState stOs = eUnspecified;
+
+      PRBool badFlag = PR_FALSE;
+
+      while (nsnull != (token = nsCRT::strtok(whitespace, kWhitespace, &whitespace)) &&
+             !badFlag) {
+        NS_ConvertASCIItoUTF16 wtoken(token);
+        ToLowerCase(wtoken);
+
+        if (CheckStringFlag(kApplication, wtoken, appID, stApp) ||
+            CheckStringFlag(kOs, wtoken, osTarget, stOs) ||
+            CheckVersionFlag(kOsVersion, wtoken, osVersion, vc, stOsVersion) ||
+            CheckVersionFlag(kAppVersion, wtoken, appVersion, vc, stAppVersion))
+          continue;
+
+        LogMessageWithContext(manifestURI, line, nsIScriptError::warningFlag,
+                              "Warning: Unrecognized chrome registration modifier '%s'.",
+                              token);
+        badFlag = PR_TRUE;
+      }
+
+      if (badFlag || stApp == eBad || stAppVersion == eBad || 
+          stOs == eBad || stOsVersion == eBad)
+        continue;
+      
+      nsDependentCString host(package);
+
+      PRBool exists;
+      rv = rph->HasSubstitution(host, &exists);
+      NS_ENSURE_SUCCESS(rv, rv);
+      if (exists) {
+        LogMessageWithContext(manifestURI, line, nsIScriptError::warningFlag,
+                              "Warning: Duplicate resource declaration for '%s' ignored.",
+                              package);
+        continue;
+      }
+
+      nsCOMPtr<nsIURI> resolved;
+      rv = io->NewURI(nsDependentCString(uri), nsnull, manifestURI,
+                      getter_AddRefs(resolved));
+      if (NS_FAILED(rv))
+        continue;
+
+      rv = rph->SetSubstitution(host, resolved);
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
     else {
       LogMessageWithContext(manifestURI, line, nsIScriptError::warningFlag,
                             "Warning: Ignoring unrecognized chrome manifest instruction.");
