@@ -57,7 +57,7 @@ function initMunger()
      * - end with whitespace, non-word, or end-of-line
      */
     client.linkRE =
-        /(?:\s|\W|^)((?:(\w[\w-]+):[^\s]+|www(\.[^.\s]+){2,})\b[\/=]?)(?=\s|\W|$)/;
+        /(?:\s|\W|^)((?:(\w[\w-]+):[^\s]+|www(\.[^.\s]+){2,})\b[\/=\)]?)(?=\s|\W|$)/;
 
     // Colours: \x03, with optional foreground and background colours
     client.colorRE = /(\x03((\d{1,2})(,\d{1,2}|)|))/;
@@ -149,11 +149,21 @@ function insertLink(matchText, containerTag, data, mungerEntry)
     var linkText;
 
     var trailing;
-    ary = matchText.match(/([.,?]+)$/);
+    ary = matchText.match(/([.,?\)]+)$/);
     if (ary)
     {
         linkText = RegExp.leftContext;
         trailing = ary[1];
+        
+        // We special-case links that end with (something), often found on wikis
+        // if "trailing" starts with ) and there's an unclosed ( in the
+        // "linkText"; then we put the final ) back in
+        if ((trailing.indexOf(")") == 0) && (linkText.match(/\([^\)]*$/)))
+        {
+            
+            linkText += ")";
+            trailing = trailing.substr(1);
+        }
     }
     else
     {
@@ -205,8 +215,7 @@ function insertLink(matchText, containerTag, data, mungerEntry)
         client.prefs["urls.list"].update();
     }
 
-    var anchor = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                          "html:a");
+    var anchor = document.createElementNS(XHTML_NS, "html:a");
     var mircRE = /\x1f|\x02|\x0f|\x16|\x03([0-9]{1,2}(,[0-9]{1,2})?)?/g;
     anchor.setAttribute("href", href.replace(mircRE, ""));
 
@@ -246,8 +255,7 @@ function insertMailToLink(matchText, containerTag, eventData, mungerEntry)
     else
         href = matchText;
 
-    var anchor = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                          "html:a");
+    var anchor = document.createElementNS(XHTML_NS, "html:a");
     var mircRE = /\x1f|\x02|\x0f|\x16|\x03([0-9]{1,2}(,[0-9]{1,2})?)?/g;
     anchor.setAttribute("href", href.replace(mircRE, ""));
 
@@ -290,8 +298,7 @@ function insertChannelLink(matchText, containerTag, eventData, mungerEntry)
 
     var linkText = removeColorCodes(matchText);
     var encodedLinkText = fromUnicode(linkText, eventData.sourceObject);
-    var anchor = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                          "html:a");
+    var anchor = document.createElementNS(XHTML_NS, "html:a");
     anchor.setAttribute("href", eventData.network.getURL(encodedLinkText));
 
     // Carry over formatting.
@@ -319,8 +326,7 @@ function insertTalkbackLink(matchText, containerTag, eventData, mungerEntry)
         return;
     }
 
-    var anchor = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                          "html:a");
+    var anchor = document.createElementNS(XHTML_NS, "html:a");
 
     anchor.setAttribute("href", "http://talkback-public.mozilla.org/" +
                         "search/start.jsp?search=2&type=iid&id=" + matchText);
@@ -382,8 +388,7 @@ function insertBugzillaLink (matchText, containerTag, eventData, mungerEntry)
             }
         }
 
-        var anchor = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                              "html:a");
+        var anchor = document.createElementNS(XHTML_NS, "html:a");
         anchor.setAttribute("href", bugURL);
         // Carry over formatting.
         var otherFormatting = calcClass(eventData);
@@ -418,8 +423,7 @@ function insertRheet(matchText, containerTag, eventData, mungerEntry)
         return;
     }
 
-    var anchor = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                          "html:a");
+    var anchor = document.createElementNS(XHTML_NS, "html:a");
     anchor.setAttribute("href",
                         "http://ftp.mozilla.org/pub/mozilla.org/mozilla/libraries/bonus-tracks/rheet.wav");
     anchor.setAttribute("class", "chatzilla-rheet chatzilla-link");
@@ -434,6 +438,7 @@ function insertQuote (matchText, containerTag)
         containerTag.appendChild(document.createTextNode("\u201c"));
     else
         containerTag.appendChild(document.createTextNode("\u201d"));
+    containerTag.appendChild(document.createElementNS(XHTML_NS, "html:wbr"));
 }
 
 function insertSmiley(emoticon, containerTag)
@@ -491,8 +496,7 @@ function insertSmiley(emoticon, containerTag)
         return;
     }
 
-    var span = document.createElementNS ("http://www.w3.org/1999/xhtml",
-                                         "html:span");
+    var span = document.createElementNS(XHTML_NS, "html:span");
 
     /* create a span to hold the emoticon text */
     span.setAttribute ("class", "chatzilla-emote-txt");
@@ -506,8 +510,7 @@ function insertSmiley(emoticon, containerTag)
      * to turn off the emoticon text, but keep the image.  ie.
      * chatzilla-emote-txt { display: none; } turns off
      * chatzilla-emote-txt:after as well.*/
-    span = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                    "html:span");
+    span = document.createElementNS(XHTML_NS, "html:span");
     span.setAttribute("class", "chatzilla-emote");
     span.setAttribute("type", type);
     span.setAttribute("title", emoticon);
@@ -633,9 +636,8 @@ function mircReverseColor (text, containerTag, data)
 
 function showCtrlChar(c, containerTag)
 {
-    var span = document.createElementNS ("http://www.w3.org/1999/xhtml",
-                                         "html:span");
-    span.setAttribute ("class", "chatzilla-control-char");
+    var span = document.createElementNS(XHTML_NS, "html:span");
+    span.setAttribute("class", "chatzilla-control-char");
     if (c == "\t")
     {
         containerTag.appendChild(document.createTextNode(c));
@@ -645,8 +647,9 @@ function showCtrlChar(c, containerTag)
     var ctrlStr = c.charCodeAt(0).toString(16);
     if (ctrlStr.length < 2)
         ctrlStr = "0" + ctrlStr;
-    span.appendChild (document.createTextNode ("0x" + ctrlStr));
-    containerTag.appendChild (span);
+    span.appendChild(document.createTextNode("0x" + ctrlStr));
+    containerTag.appendChild(span);
+    containerTag.appendChild(document.createElementNS(XHTML_NS, "html:wbr"));
 }
 
 function insertHyphenatedWord(longWord, containerTag, data)
@@ -656,13 +659,12 @@ function insertHyphenatedWord(longWord, containerTag, data)
     if (data && ("hasColorInfo" in data))
         newClass = calcClass(data);
 
-    var wbr = document.createElementNS("http://www.w3.org/1999/xhtml",
-                                       "html:wbr");
+    var wbr = document.createElementNS(XHTML_NS, "html:wbr");
     for (var i = 0; i < wordParts.length; ++i)
     {
         if (newClass)
         {
-            var newTag = document.createElementNS(NS_XHTML, "html:span");
+            var newTag = document.createElementNS(XHTML_NS, "html:span");
             newTag.setAttribute("class", newClass);
             newTag.appendChild(document.createTextNode(wordParts[i]));
             containerTag.appendChild(newTag);
@@ -690,7 +692,7 @@ function insertInlineButton(text, containerTag, data)
     var title = ary[2];
     var command = ary[3];
 
-    var link = document.createElementNS("http://www.w3.org/1999/xhtml", "a");
+    var link = document.createElementNS(XHTML_NS, "html:a");
     link.setAttribute("href", "x-cz-command:" + encodeURI(command));
     link.setAttribute("title", title);
     link.setAttribute("class", "chatzilla-link");
